/*!
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */svg{overflow:hidden}.node rect{stroke:#51504f;stroke-width:1px;fill:#fff;transition:stroke .2s ease-in-out,opacity .2s ease-in-out}.node rect[data-highlight=highlight]{stroke:#017cee!important}.edgeLabel{transition:opacity .2s ease-in-out}.edgeLabel>.label{font-size:10px;font-weight:400}.edgeLabel text{transform:translate(16px)}.edgeLabel rect{fill:#fff}.edgePath{stroke:#51504f;stroke-width:1px;fill:none;transition:stroke .2s ease-in-out,opacity .2s ease-in-out}.edgeLabel[data-highlight=fade],.edgePath[data-highlight=fade],.edgePaths[data-highlight=fade]>.edgePath{opacity:.2!important}.edgePath[data-highlight=highlight]{stroke:#017cee}.edgePath .arrowhead{stroke:none!important;fill:#51504f;stroke-width:0!important;transition:fill .2s ease-in-out,opacity .2s ease-in-out}.edgePath[data-highlight=highlight] .arrowhead{stroke:#017cee;fill:#017cee}.not-allowed rect,.not-allowed text{cursor:not-allowed!important}g.cluster rect{stroke:#fff;stroke-dasharray:5;rx:5;ry:5;opacity:.5}g.node{transition:opacity .2s ease-in-out}g.node[data-highlight=fade]{opacity:.2!important}g.node rect{stroke:#fff;stroke-width:1.5px;cursor:pointer}g.node circle{stroke:#51504f;stroke-width:1.5px;cursor:pointer}g.node .label{font-size:.7em;font-weight:400;pointer-events:none}g.node text{cursor:pointer}.svg-wrapper{background-color:#f0f0f0;border-radius:4px;cursor:move;height:100%;overflow:hidden;width:100%}#graph-svg{overflow:visible}.refresh-actions{align-items:center;background-color:#f0f0f0ee;display:inline-flex;justify-content:flex-end;margin-bottom:15px;margin-top:10px;min-width:225px;position:absolute;right:20px}.legend-item.dag{background-color:#e8f7e4;float:left}.legend-item.trigger{background-color:#ffefeb;float:left}.legend-item.sensor{background-color:#e6f1f2;float:left}.legend-item.dataset{background-color:#fcecd4;float:left}g.node.dag rect{fill:#e8f7e4}g.node.trigger rect{fill:#ffefeb}g.node.sensor rect{fill:#e6f1f2}g.node.dataset rect{fill:#fcecd4}