/*!
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

/* global describe, test, expect, jest, window */

import React from "react";
import "@testing-library/jest-dom";
import { render, fireEvent } from "@testing-library/react";

import { ClipboardButton } from "./Clipboard";

describe("ClipboardButton", () => {
  test("Loads button", async () => {
    const windowPrompt = window.prompt;
    window.prompt = jest.fn();
    const { getByText } = render(<ClipboardButton value="lorem ipsum" />);

    const button = getByText(/copy/i);
    fireEvent.click(button);
    expect(window.prompt).toHaveBeenCalledWith(
      "Copy to clipboard: Ctrl+C, Enter",
      "lorem ipsum"
    );
    window.prompt = windowPrompt;
  });
});
