/*!
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

.panel-heading #alerts-accordion-toggle::after {
  float: right;
  color: grey;
}

.panel-heading #alerts-accordion-toggle.collapsed::after {
  float: right;
  color: grey;
}

#errorHeading {
  background-color: #d6d8d9; /* same color as Bootstrap's Dark Alert */
}

.dag-import-error {
  position: relative;
  white-space: pre;
  height: 14px;
  line-height: 14px; /* show only one line of 14px text */
  overflow: hidden;
  cursor: pointer;
}

.expanded-error {
  height: 100%;
}

.dag-import-error::after {
  float: right;
  color: #e43921;
  position: absolute;
  top: 0;
  right: 0;
}

.expanded-error .toggle-direction {
  transform: rotate(180deg);
}
