try{module.exports.SwaggerUIBundle=require("./swagger-ui-bundle.js"),module.exports.SwaggerUIStandalonePreset=require("./swagger-ui-standalone-preset.js")}catch(e){}module.exports.absolutePath=require("./absolute-path.js"),module.exports.getAbsoluteFSPath=require("./absolute-path.js");