/*! For license information please see durationChart.e988d20fb6cfb21ab0fd.js.LICENSE.txt */
!function(e,o){"object"==typeof exports&&"object"==typeof module?module.exports=o():"function"==typeof define&&define.amd?define([],o):"object"==typeof exports?exports.Airflow=o():(e.Airflow=e.Airflow||{},e.Airflow.durationChart=o())}(self,(function(){return function(){function e(){$("#isCumulative").is(":checked")?($("#dur_chart").hide(),$("#cum_dur_chart").show()):($("#dur_chart").show(),$("#cum_dur_chart").hide())}return $(document).on("chartload",e),$(document).ready(e),$("#isCumulative").on("click",e),{}}()}));